#!/usr/bin/perl
# (c) napp-it  2012-2024 replicate cs # Bookmarks: 0,0 0,0 0,0 0,0 0,0 0,0 0,0 0,0 0,0 0,377
# apr 1, 2024

# debug: run from cli    perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/job-replicate.pl run_id
# auto.pl:    `powershell start-process  -WorkingDirectory $wpath -filepath C:\\xampp\\perl\\bin\\perl.exe -argumentlist \"$t\",\"run_$id\" -WindowStyle minimized`;       # minimized=debugmodr
# this script always runs locally on cs client (Windows) and controls destination system remotely

# how to
# id.running is created first and blocks restarts

# replications are "remote" via netcat or local zfs send | zfs receive
# all receiver pid            -> job.running
# all remote sender pid       -> job.sender
# progress in % (fsdiff)      -> job.progress

# job can be started manually or via auto.pl -> powerscript create-process
# or manuall at console via perl script.pl run_jobid

#job par
#############################
print <<EoF;
# parameter
#job=replicate
#source_ip=192.168.2.203
#destination_ip=192.168.2.75
#source_zfs=daten1/test
#destination_zfs=bsd/data/1
#writeable=yes
#zfs_name=test
#transport=nc
#port=52969
#recursiv=no
#send_raw=no
#esxi_ip=
#send_i=-i
#send_options=
#rec_options=
#pv_limit=
#keep=
#hold=
#month=every
#day=sun
#hour=0
#min=0
#status=manual
#last=new
##########################


EoF

##################
# netcat
# send:    zfs send snap | nc ip 7000
# receive: nc -l -d -w 20 | zfs receive tank/data
##################

# ok linux
# zfs send  tank/data@2024.03.17.18.34.15 | nc -w 20 127.0.0.1 7000
# nc -l -p 7000 | zfs receive -F  tank/data/new

# test all platforms for wait in listen and send
# -d (detach, runs netcat in the background)

# nc -l -w 5 8080          ->   listen 5s. End when no connect in listen mode
# nc -w 5 server.com 80

### special settings??
# bsd
# remote send ok when receive is waiting
# zfs send tank/data@nc | nc -w 10 192.168.2.71 7000
# local nc transfer: no, must use send | receive

# linux (127.0.0.1 not allowed, must use regular ip)
# local and remote nc ok
# ok: zfs send tank/data3@nc | nc  -w 10 192.168.2.71 7000 | nc -l  -p 7000 -w 10 | zfs receive -Fv tank/data4

# osx

# solaris/illumos

# windows:
#################
# -d      detach from stdin
# -p      port
# -c      send crlf instead lf
# nc -l -p port [options]        # listen
# nc [options] hostname  port    # send

# listen: nc -l -d -w 20
# send:   nc -w 20 ip 7000


#use strict;                                                   # disable for syntax check
use Fcntl qw(:flock);                                          # filelocking
$|=1;                                                          # keine Pufferung, no buffering

use vars qw($t1 $on $wpath $dpath $tpath $par $id $action $f $t @t $p @p %current $error $info %job $buf $os %sys $pids @keeph @keepd @keepm @keepy $keepfs $holdnr $debug $ncwait);

$wpath="/xampp/web-gui";
$dpath="/xampp/web-gui/data";
$tpath="/xampp/web-gui/tmp";

# netcat buffer size on sol
$buf="131072";
$nc="nc";                    # replaced by server.pl
$ncwait=60;  # default

# common lib
require "$dpath/menues/_lib/windows/scripts/task-lib.pl";        #
require "$dpath/menues/_lib/windows/grouplib.pl";
require "$dpath/wwwroot/cgi-bin/admin-wlib.pl";

$par=$ARGV[0];
if (!(defined($par))) {
     #$t=`cat /var/web-gui/_log/napp-it.cfg | grep nc_wait`;   chomp($t);
     $t=my_read("$wpath/_log/napp-it.cfg","nc_wait");          # read and filter
     $t=~s/\D//gs;
     if ($t>10) { $ncwait=$t; }
     &log("job-replication |  missing parameter","m");
     print __LINE__. "job-replication: missing parameter ex run_1234567 or run_1234567_12345\n";
     exit;
}

 #debug=012 etc
 $debug=$ARGV[1];                                  # ex 3=keep debug
 if ($debug > 10000) { $debug=""; }                # debug mode ex 123 or time()

 #par: action_job timestamp
 $par=~s/ .*//;
 chomp($par);

 # parameter: action_$job time
 if (!($par =~/\w+\_\d+$/)) {
     print "job-replication 57: wrong par $par\n";
      print __LINE__." wrong par\n";
     exit;
 }
 @t=();
 @t=split(/\_/,$par);    # action job
 $action=$t[0];
 $action=~s/[^a-z]//g;

 $id=$t[1];
 chomp($id);
 #############################################

# run?
if ($action eq "")                     {  &my_log_last("wrong action - must be not empty");     exit; }
if (!("run,cancel,delete"=~/$action/)) {  &my_log_last("wrong action - must be run: $action");  exit; }

#cancel or delete
if ($action eq "cancel") { &my_cancel; $info="cancel job"; &my_fast_end; exit; }
if ($action eq "delete") { &my_delete; exit; }                                       # call cancel, log, exit in sub

# get par -> %current
 &task_read_jobpar($id);
 &my_par($id);

 # !! if destination is 127.0.0.1 !!
 # replace 127.0.0.1 with ip as seen from source host via &socket("ip",sourcehost)

 $job{'ncsend_ip'}=$job{'destination_ip'};                     # default, use for nc send
 if ($job{'destination_ip'} eq "127.0.0.1") {
      $job{'ncsend_ip'}=&socket('ip',$job{'source_ip'});       # 127.0.0.1 ip from sender view
      if ($job{'ncsend_ip'}=~/error/)  { &my_fast_end;   exit; }
      &my_log_id("replace send  127.0.0.1 -> send $job{'ncsend_ip'}");
      print "169 replace send 127.0.0.1 -> send $job{'ncsend_ip'}\n";
}



if ($action eq "run") {
     $t=&my_job_lock;

     if ($t ne "ok") {
         print "cannot lock replication job $id: $t\n";
         &my_log_id("cancel next, job is running");                           # only last.log to preserver error as default
         &my_log_last("next replication cancelled<br>job is running");
         exit;
     } else {
         print "start next replication $id\n";
         &my_log_id("next replication started");
     }
}

# log par
  $t=&print_my_par($id);  $t=~s/\n/<br>/gs;
  &my_log_last("90,par$t");

 #debug


#start detailed log
 &task_gettime;
 $t="$wpath/_log/jobs/$id.last";
    open (PF, ">$t");
      flock(PF, LOCK_EX);
      print PF "$current{'timestamp'}\tmain line 76\tstart job-replicate with parameter:<br>id=$id, action=$action, par=$par\n";
    close (PF);

# logfile receiver started
 if (-f "$tpath/$id.receiver") { unlink("$tpath/$id.receiver"); }        # opt

 # opt pre job
 #if (-f "$wpath/_log/jobs/$id.pre") {
 #      print "#### 149 todo: pre job\n";
 #      &my_log_last("(149) execute powershell script $id.pre");
 #      #$t=`sudo sh /var/web-gui/_log/jobs/$id.pre`;             # start a local win process
 #      &my_log_last("result of pre script\: $t");
 #}

  # ESXi snaps
  #if ($job{'esxi_ip'} =~/\d/) {
  #  print "#### 157 todo:esxi job\n";
  #  #$t=`perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/esxi/job-esxisnap.pl create_$id`;   # create ESXi snaps, todo
  #}

 # start replication
 &my_run;
 exit;


######## run job (child process)
sub my_run {
###########

    # check if dest pool is mounted
    my $pool=$job{'destination_zfs'};  # pool/fs
    $pool=~s/\/.*//;                   # pool
    $r = &socket("zfs list -H -o name $pool",$job{'destination_ip'},120);
    print __LINE__." ip $job{'destination_ip'} check dest pool -> zfs list -H -o name $pool\n$r\n";
    if ($r=~/error/)  {  &my_log_last("check dest pool: $r"); &my_fast_end; exit; }

    if ($r=~/not exist/s)  {
        print __LINE__." pool $r not mounted\n";
        &my_log_last("error","pool $r not mounted");
        &my_log_id("error","pool $r not mounted");
        &my_fast_end; exit;
    }

    # check if dest fs exists
    $r = &socket("zfs list -o name | grep $job{'destination_zfs'}","$job{'destination_ip'}");
    if ($r=~/error/)  { &my_log_last("check dest fs: $r");  &my_fast_end;  exit; }

    &my_log_last("check if dest fs exists: $r");
    @t=split(/\n/,$r);
    $r="";
    foreach my $t (@t) {
        if ($t eq $job{'destination_zfs'}) { $r="$job{'destination_zfs'}"; }      # fs there, only incr. repli possible
    }

    #debug
    #print "$job{'host'}\n$job{'hostip'}\n"; exit;

    # remote netcat, locally send | receive
    ################
    if ($r eq "") {  # initial
        if ($job{'source_ip'} eq $job{'destination_ip'})  {
              print __LINE__." local initial replication\n";
              &log("$id <br>job-replicate 2246 | start replication<br>local initial | $job{'source_ip'} -> $job{'destination_ip'}<br>$job{'source_zfs'} -> $job{'destination_zfs'} | ","m");     #monitor
              &my_local_initial_replication;

        } else {         # remote initial
              print __LINE__." remote nc initial replication\n";
              &log("$id <br>job-replicate 2246 | start replication<br>remote initial | $job{'source_ip'} -> $job{'destination_ip'}<br>$job{'source_zfs'} -> $job{'destination_zfs'} | ","m");     #monitor
              &my_remote_initial_replication;

         }

    } else {  # delta/incr

        if ($job{'source_ip'} eq $job{'destination_ip'})  {
              print __LINE__."  local incremental replication\n";
              &log("$id <br>job-replicate 255 | start replication<br>local incremental | $job{'source_ip'} -> $job{'destination_ip'}<br>$job{'source_zfs'} -> $job{'destination_zfs'} | ","m");     #monitor
              &my_local_delta_replication;

        } else {
              print __LINE__." remote nc incremental replication\n";
              &log("$id <br>job-replicate 255 | start replication<br>remote incremental | $job{'source_ip'} -> $job{'destination_ip'}<br>$job{'source_zfs'} -> $job{'destination_zfs'} | ","m");     #monitor
              &my_remote_delta_replication;

         }
    }

     exit;
 }



 ##############
 sub my_job_lock {
 #############

    #
    # a job should not restart while running
    # a running replication job is indicated by a file jobs/$id.running
    # with the next destination snap as content.

    # howto
    # 1. return "running" when there is a file $id.running

    # or
    # 2. create a file $id.running and lock it to avoid restarts
    # 3. determine next destinatination snap number and write it to the file
    #   if this snap is created, this is an indication of a successful run
    # 4. when job runs ok, delete $id.running

    my $running="$wpath/_log/jobs/$id.running";
    my $lock="$wpath/_log/jobs/$id.lock";

   # debug
    print &print_my_par;

    if (!($id=~/\d+/)) {  return ("error: missing id"); }
    if (-f "$running") {  return ("running"); }

    open (X,"<$lock");                                 # lock this id
          flock(X, LOCK_EX);
          #
          &my_write("$running","1");                   # create id.running to block restarts

    close (X);
    return('ok');
}



#############
sub my_lopt {
############  netcat listen options

        # listen options
        # solaris needs -d -l on listen
        # bsd does not want -p port, just port

        my $r=$_[0];
        my $lopt =" -w 20 ";          # default

        if ($r =~/linux/i)           {  $lopt="-w 20 -p";  }       # port follows
        if ($r =~/mswin/i)           {  $lopt="-d w 20 -p" ; }
        if ($r =~/solaris/i)         {  $lopt="-d -w 20 -p" ; }
        if ($r =~/darwin/i)          {  $lopt=" -w 20";  }    # osx
        if ($r =~/bsd/i)             {  $lopt="-d";           }    # on bsd: do not use -w or -p 123, just 123
        return($lopt);
}


#################################
sub my_remote_initial_replication {
#################################

        &my_log_last("541, start remote initial replication");
        my ($r,@v,$v,$cmd,$timer,$ok,$var,$p1,$p2,$nc,$cmd,$t);

        # ok when job_lock

        # opt delete old snap with this id (restart)
        &my_del_snap($id);          #del dest snap with id
        &update_lastrun;

        &my_log_last("start remote initial replication\n$job{source_ip} -> $job{destination_ip}");
        &my_log_id("start remote initial replication\n$job{source_ip} -> $job{destination_ip}");

        $r=&socket("zfs list -H -t filesystem,volume -o name $job{'source_zfs'}",$job{'source_ip'});
        if ($r=~/error/)  { &my_log_last("zfs list\n$r");  &my_fast_end;   exit; }

        #source fs not available??
        if (!( $r eq $job{'source_zfs'} )) {
           &my_log_last("ource zfs missing: $job{'source_zfs'}");
           &my_fast_end;  exit;
        }

        #set receiver ip
        $job{'myip'}=$job{'destination_ip'};
        if (!($job{'myip'}=~/^\d+\.\d+/)) {  $info="request dest ip 287 ($job{'source_ip'} -> $job{'myip'})"; $error="error";  &my_fast_end; exit;  }


        #check source snaps
        #$r = &grouplib_ask_remote_snaplist("$job{'hostip'}");
        $r=&socket("zfs list -H -r -t snap -o name $job{'source_zfs'}",$job{'source_ip'});
        if ($r=~/error/)  {   &my_log_last("zfs list\n$r");  &my_fast_end;  exit; }

        my @rsnaps=();
        @rsnaps=split(/,/,$r);                                                             # commalist !!

        # opt delete all source snaps in case of restart
        foreach my $s (@rsnaps) {
              if (!($s=~/^$job{'$source_zfs'}.*\@$id\_repli/)) { next; }                       # inkl recursiv snaps
              #$t=&grouplib_ask_remote_destroy_snap("$job{'hostip'}","snap_name=$s");

              print "632 destroy snap $job{'source_ip'} $s\n";
              &socket("zfs destroy $s",$job{'source_ip'});
              if ($r=~/error/)  {   &my_log_last("zfs destroy $s\n$r");  &my_fast_end;  exit; }
        }

        #start receiver
        $nc="nc";                # default os

        $r=&socket("on",$job{'destination_ip'});
        if ($r=~/error/)  {   &my_log_last("on");  &my_fast_end;  exit; }

        my $lopt=&my_lopt($r);      # listen options

        # receiver cmd
        $cmd="l_$id $nc -l $lopt $job{'port'} | zfs receive $job{'rec_options'} -Fv $job{'destination_zfs'} &";              # "l_123 command
        $cmd=~s/ +/ /g;

        &my_log_id("661, $job{'destination_ip'} start receiver:","$cmd");
        &my_log_last("672,$job{'destination_ip'} start receiver:","$cmd");
        &log("$id <br>job-replicate 63 | start receiver $job{'destination_ip'}| $cmd |  ","m");

        # start receiver on destination
        print __LINE__." $job{'destination_ip'} start initial receiver:\n$cmd\n\n";
        $t1=time();                     # start time

        #################
        ## start receiver
        #################

        print "\n";    # nl
        $current{'receiver_pids'}=&socket($cmd,$job{'destination_ip'});         # start receiver
        if ($current{'receiver_pids'}=~/error/)  {  &my_fast_end;  exit; }

        &my_log_id("receiver $job{'destination_ip'} pids","$current{'receiver_pids'}");
        &my_write("$wpath/_log/jobs/$id.receiver",$current{'receiver_pids'});    # pid list of receiver tasks


        sleep(3);
        ###########################
        #verify remote receiver cmd
        ###########################
        $r=&socket("rl",$job{'destination_ip'});            # remote log
        if ($r=~/error/)  {   &my_log_last("remote log\n$r");  &my_fast_end;  exit; }

        @t=split(/\n/,$r);
        @t=reverse(@t);

        #  /xampp/web-gui/data/cs_server/tools/nc/win/nc64 -w 30 -l -p 58527 | zfs receive -Fv tank/data4 2>&1 >>/xampp/web-gui/data/cs_server/tmp/remote.log &
        foreach my $t (@t) {
             if (($t=~/$job{'port'}/) && ($t=~/zfs receive/))    {
                  &my_log_id("receive cmd on $job{'destination_ip'}","$t");
                  last;
             }
        }

        ############
        ## sender ##
        ############

        sleep(5);

        # start sender
        print "\n";
        #  zfs send  snap@ | nc -w 20 127.0.0.1 7000

        # create initial repli snap 1
        # snap_name1=tank/data@1354303765_repli_zfs_192.168.5.6_nr_1
        # tank/data is source fs, 134.. jobid and ip is destination ip

        $t="$job{'source_zfs'}\@$id\_repli_zfs_$job{'source_ip'}\_nr_1";
        $nc="nc";

        #initial repli snap 1
        # snap_name1=tank/data@1354303765_repli_zfs_192.168.5.6_nr_1

        # opt del (failed initial)
        $r=&socket("zfs list -H -r -o name -t snapshot $t",$job{'source_ip'});
        #if ($r=~/error/)  {  &my_fast_end;  }

        if ($r eq $t)  {
               print "702, zfs destroy $t\n";
               $r=&socket("zfs destroy $t",$job{'source_ip'});
                if ($r=~/error/)  {   &my_log_last("zfs destroy $t\n$r");  &my_fast_end;  exit; }
        }

        # create send snap
        print __LINE__." $job{'source_ip'}, zfs snapshot $t\n";
        &my_log_id("send snap from $job{'source_ip'}","$t");
        &my_log_last("send snap from $job{'source_ip'}","$t");

        $r=&socket("zfs snapshot $t",$job{'source_ip'});
        if ($r=~/error/)  {   &my_log_last("zfs snapshot $t\n$r");  &my_fast_end;  exit; }

        # send
        if ($job{'send_raw'}=~/yes|true|1/) {
          $job{'send_raw'}="--raw";
        }  else {
              $job{'send_raw'}="";
        }

        if ($job{'recursiv'}=~/yes|true|1/) {
               $job{'send_rec'}="-R";
        }  else {
              $job{'send_rec'}="";
        }

        # replace dest_ip with ncsend_ip for dest 127.0.0.1
        $cmd="l_$id zfs send  $job{'send_rec'} $job{'send_raw'} $job{'send_options'} $t | nc -w 30 $job{'ncsend_ip'} $job{'port'} &";   $cmd=~s/ +/ /g;

        &my_log_id("start initial send to $job{'ncsend_ip'}","$cmd");       # keep timeout error state until ok
        &my_log_last("start initial send to $job{'ncsend_ip'}","\n$cmd");

        # send
        print "\n";    # nl
        $current{'sender_pids'}=&socket("$cmd",$job{'source_ip'});
        if ($current{'sender_pids'}=~/error/)  { &my_log_last("sender pids","$current{'sender_pids'}");  &my_fast_end;  exit; }


        sleep(2);
        ###########################
        #verify remote sender cmd
        ###########################
        $r=&socket("rl",$job{'source_ip'});            # remote log
        if ($r=~/error/) {   &my_fast_end;  exit;  }

        @t=split(/\n/,$r);
        @t=reverse(@t);

        #  zfs send tank/data@1711788007_repli_zfs_192.168.2.75_nr_1 | /var/web-gui/cs_server/tools/nc/bsd/nc -w 30 192.168.2.76 58527 2>&1 >>/var/web-gui/cs_server/tmp/remote.log &
        foreach my $t (@t) {
             if (($t=~/$job{'port'}/) && ($t=~/zfs send/))    {
                  &my_log_id("receive cmd on $job{'source_ip'}","$t");
                  print __LINE__." $job{'source_ip'}: $t\n";
                  last;
             }
        }


        ##########################
        # wait for receiver to end
        ##########################

        $t1=time();

        # get size of source fs
        my $size=&socket("zfs list -H -t filesystem,volume -o used $job{'source_zfs'}",$job{'source_ip'});
         if ($size=~/error/)  {   &my_log_last("zfs list\n$size");  &my_fast_end;  exit; }

        # job zfs-source size needed in monitor
        $size=&my_val2mb($size);
        $r="$size\_";
        &my_write("$wpath/_log/jobs/$id.progress",$r);

        # set timeout error as default
        &my_log_timeout;

        # ended when next_dest_snap is created + update progress
        my $next_dest_snap ="$job{'destination_zfs'}\@$id\_repli_zfs_$job{'source_ip'}\_nr_1";
        my $dsize;         # destination size
        my $last_dsize=0;  # last destination size
        my $percent;       # percent done
        my $timeout=0;

        my $dest_parent=$job{'destination_zfs'};
        $t=rindex($job{'destination_zfs'},"/");              # parent dest fs
        $dest_parent=substr($job{'destination_zfs'},0,$t);
        print "751 dest parent: $dest_parent\n";
        my ($dest_snaps,$dest_fs);

        print "\n";

        while (1) {
             sleep(40);                # check after/once every 40s

              # next dest snap created?
              $dest_snaps=&socket("zfs list -r -H -o name -t snapshot $dest_parent",$job{'destination_ip'},120);     #timeout 120s
              if ($dest_snaps=~/error/)  {  &my_fast_end;  exit; }

              $dest_fs=&socket("zfs list -r -H -o name -t filesystem $dest_parent",$job{'destination_ip'},120);     #timeout 120s
              if ($dest_sf=~/error/)  {  &my_fast_end;  exit; }

              if ($dest_snaps=~/$next_dest_snap/s) {

                  &my_log_last("next dest snap: $next_dest_snap is created");
                  last;

              } else {

                  # update progress, source size in $size   id.progress:  sourcesize_destsize_percent
                  # get size of dest fs

                  if ($dest_fs=~/$job{'destination_zfs'}/s) {
                        #
                        $dsize=&socket("zfs list -H -t filesystem,volume -o used $job{'destination_zfs'}",$job{'destination_ip'});
                        if ($dsize=~/error/)  { &my_fast_end;  exit; }

                        $dsize=&my_val2mb($dsize);

                        if ($size > 0) {

                           $percent=int($dsize/$size*100);  # 3.1 -> 3
                           &my_write("$wpath/_log/jobs/$id.progress","$size\_$dsize\_$percent");


                           if ($dsize > $last_dsize) {
                                 $timeout=0;         # reset timeout
                                 &my_log_last("wait until next dest snap is created: $next_dest_snap, $percent % done");
                           }  else {
                                print __LINE__." wait until next dest snap is created: $next_dest_snap, $percent % done\n";
                           }
                        }
                    }  else {
                         print __LINE__." wait for creation of dest fs $job{'destination_zfs'}\n";

                         $r=&socket("ps axw | grep zfs",$job{'destination_ip'});
                         if ($r=~/error/)  { &my_fast_end;  exit; }

                         $t=&socket("rl | grep receiving",$job{'destination_ip'});
                         if ($t=~/error/)  { &my_fast_end;  exit; }
                         $t=~s/.*receiving/receiving/s;     # only last

                         &my_log_last("processes on destination $job{'destination_ip'}\n$r\n$t");


                    }
              }

              $last_dsize=$dsize;

              $timeout++;         # 80s
              if ($timeout > 1) {
                 &my_log_id("timeout without any data received");
                 &my_log_last("timeout without any data received");   # prints on console
                 &my_fast_end;  exit;
              }
        }

        #################### /end waiting

        ###################
        ## end with success
        ###################

        &my_ok_and_exit;

}


#################################
sub my_remote_delta_replication {
#################################

    my ($r,@v,$v,$cmd,$timer,$ok,$var,$p1,$p2,$nc,@s,$s,$dnr,$dnr2,$dnr3,$t,$used);

     $ncwait=60;
     $nc="nc";
     &update_lastrun;

     &my_log_last("start remote incremental replication\n$job{source_ip} -> $job{destination_ip}");
     &my_log_id("start remote incremental replication\n$job{source_ip} -> $job{destination_ip}");

     #step 1 check for newest source dest snaps
     my $newest_src="0";
     my $newest_dest="0";

     my $newest_src_snap="";
     my $newest_dest_snap="";
     my $next_src_snap="";
     my $next_dest_snap="";

     #size of source
     $r=&socket("zfs list -r -H -o name | grep $job{'source_zfs'}",$job{'source_ip'});
     if ($r=~/error/)  {  &my_log_last("zfs list\n$r"); &my_fast_end;  exit; }

     &my_log_last("src fs on $job{'source_ip'}\n$r");
     $r=~s/\n/,/gs;

     #source fs not available??, lines separated by ,
     $t=" $r ";  $t=~s/,/ /gs;

     if (!($t=~/ $job{'source_zfs'} /s)) {
        &my_log_last("source fs missing: $job{'source_zfs'}");
        &my_fast_end; exit;
     }

     # source snaps
     $r=&socket("zfs list -H -r -o name -t snapshot | grep $id",$job{'source_ip'});
     if ($r=~/error/)  {  &my_log_last("zfs list\n$r"); &my_fast_end;  exit; }
     $r=~s/\n/,/gs;

     @s=split(/,/,$r);

     ($newest_src_snap,$newest_dest_snap,$next_src_snap,$next_dest_snap) =&my_newest_snap_pair(@s);  # remote transfer, add remote zfs

     $current{'next_dest_snap'}=$next_dest_snap;
     # last        = last dest prior repli run
     # newest_dest = current base snap
     # next_dest   = new snap after repli run

     #print "977, newestsrc=$newest_src_snap\nnewest_dest=$newest_dest_snap\nnewest_src=$next_src_snap\n$next_dest=next_dest_snap\n";

     # check remote snaps for current nr
     $ok="0";
     foreach my $s (@s) {
          print __LINE__." $s\n";
          if ($s eq "$newest_src_snap") {
              $ok="1";
          }
     }
     if ($ok eq "0") {
          &my_log_last("error: needed source snap missing: $newest_src_snap");
          &my_fast_end; exit;
     }


     #####
     # check remote snaps and opt delete next src snap (kept from unsuccessfull repli) or other than current
     $dnr="";
     $dnr2="";
     $dnr3="";
     if ($newest_dest_snap=~/_nr_\d+/) {
                   $dnr=$newest_dest_snap;
                   $dnr=~s/.*_nr_//;
                   $dnr2=$dnr+1;                                       # next (new) snapnr
                   $dnr3=$dnr-1;                                       # previous
     }
     if (($newest_src_snap ne "")  && ($id >0) && ($dnr >0))     {
            foreach my $s (@s) {
                       chomp($s);
                       if (!($s=~/$id\_repli/))     { next; }               # not my id

                       $s=~s/ .*//;  # remove other info but name

                       if ($s=~/_nr_$dnr2$/)          { next; }        # next nr   (keeps 3 snaps, comment for one snap)
                       if ($s=~/_nr_$dnr$/)           { next; }        # current  nr
                       if ($s=~/_nr_$dnr3$/)          { next; }        # previous  nr        # needed to destroy newest target snap and redo
                       &my_log_id("request remote snap destroy","$s");

                       $r=&socket("zfs destroy $s",$job{'source_ip'});
                       if ($r=~/error/)  {   &my_fast_end;  exit; }

            }

            $current{'newest_dest_snap'}=$newest_dest_snap;      # needed for keep
            $t=$newest_dest_snap;
            $t=~s/.*\_//;                 # keeps nr
            if ($t>1) {
                $current{'last_dest_snap'}=$newest_dest_snap;
                $current{'last_dest_snap'}=~s/\d+$//;
                $current{'last_dest_snap'}.=$t;
            }

            # do the same with  dest snaps

                $r=&socket("zfs list -H -r -o name -t snapshot -s creation | grep $job{'destination_zfs'}",$job{'destination_ip'});     # newest last
                if ($r=~/error/)  {  &my_fast_end;  exit; }

                @s=();
                @s=split(/\n/,$r);

                $t=$id."_recursiv";
                $t=$current{$t};                                             # yes
                # keep and hold does not work with recursive replication !!

                foreach my $s (@s) {
                              if (!($s=~/$id\_repli/))      { next; }        # not my snaps
                              if ($s=~/_nr_$dnr$/)          { next; }        # current  nr
                              if ($s=~/_nr_$dnr2$/)         { next; }        # next nr       (keep 3 snaps, comment for one snap)
                              if ($s=~/_nr_$dnr3$/)         { next; }        # previous nr

                              if ($s eq $last_dest_snap)    { next; }        # -2
                              if ($s eq $newest_dest_snap)  { next; }        # -1
                              if ($s eq $next_dest_snap)    { next; }        # newest

                              if ($dnr < 2) {
                                if ($s=~/\_1$|\_2$/)        { next; }        # keep on init
                              }

                              #keep snap?
                              if (&my_keep($s,$dnr)) {
                                   #print "debug 789: skip $s\n";
                                   next;
                              }

                              &my_log_id("zfs destroy","$s");

                              if ($debug=~/\w/) { print "xxx debug 793:  -- destroy $s\n"; }
                              if ($debug=~/D/) {         # Dry run
                                print "xxx debug 929:  -- would destroy $s\n";
                              } else {
                                 #dest
                                 $r=&socket("zfs destroy -f $s",$job{'destination_ip'});
                                 if ($r=~/error/)  {  &my_fast_end;  exit; }

                              }
               }
               #print "debug 802: end of snap processing\n";

     }
     ####

     #   #get my ip
     #   $job{'myip'}=&grouplib_ask_remote_get_my_ip($job{'hostip'});                # ip or name as fallback

     ################
     #receiver
     ###############

     $r=&socket("on",$job{'destination_ip'});
     if ($r=~/error/)  {   &my_log_last("on");  &my_fast_end;  exit; }

     $t1=time();

     my $lopt=&my_lopt($r);      # listen options

     $cmd="l_$id nc -l $lopt $job{'port'} | zfs receive -Fv $job{'destination_zfs'} &";   # receive
     print "start rec $cmd\n";

     &my_log_id("start receiver on $job{'destination_ip'}","$cmd");
     &my_log_last("start receiver on $job{'destination_ip'}\n$cmd");

     ## start ##
     print __LINE__." $job{'source_ip'}: zfs list -H -r -o name -t snapshot | grep $next_src_snap\n";
     $r=&socket("zfs list -H -r -o name -t snapshot | grep $next_src_snap",$job{'source_ip'});
     if ($r=~/error/)  { &my_log_last("zfs list\n$r");  &my_fast_end;  exit; }

     if ($r eq $next_src_snap)  {
               $r=&socket("zfs destroy $next_src_snap",$job{'source_ip'});
               if ($r=~/error/)  { &my_log_last("zfs destroy\n$r");  &my_fast_end;  exit; }
     }

     # create send snap
     &my_log_id("zfs snapshot $next_src_snap br>on $job{'source_ip'}: $next_src_snap");
     &my_log_last("zfs snapshot $next_src_snap <br>on $job{'source_ip'}: $next_src_snap");

     $r=&socket("zfs snapshot $next_src_snap",$job{'source_ip'});
     if ($r=~/error/)  { &my_log_last("zfs snapshot $next_src_snap\n$r");  &my_fast_end;  exit; }

     #################
     ## start receiver
     #################

     print "\n";    # nl
     $current{'receiver_pids'}=&socket($cmd,$job{'destination_ip'});         # start receiver
     if ($current{'receiver_pids'}=~/error/)  {  &my_fast_end;  exit; }

     &my_log_id("receiver on $job{'destination_ip'} started","$cmd");
     &my_log_last("receiver on $job{'destination_ip'} started","$cmd");
     &my_write("$wpath/_log/jobs/$id.receiver",$current{'receiver_pids'});    # pid list of receiver tasks

     sleep(2);
     ###########################
     #verify remote receiver cmd
     ###########################
     $r=&socket("rl",$job{'destination_ip'});            # remote log
     if ($r=~/error/) {   &my_fast_end;  exit;  }

     @t=split(/\n/,$r);
     @t=reverse(@t);

     #  /xampp/web-gui/data/cs_server/tools/nc/win/nc64 -w 30 -l -p 58527 | zfs receive -Fv tank/data4 2>&1 >>/xampp/web-gui/data/cs_server/tmp/remote.log &
     foreach my $t (@t) {
         if (($t=~/$job{'port'}/) && ($t=~/zfs receive/))    {
                  &my_log_id("receive cmd on $job{'destination_ip'}","$t");
                  print __LINE__." $job{'destination_ip'}: $t\n";
                  last;
         }
     }

     ##############
     # start sender
     ##############

     if ($job{'send_raw'}=~/yes|true|1/) {
          $job{'send_raw'}="--raw";
     }  else {
         $job{'send_raw'}="";
     }

     if ($job{'send_i'} ne "-I") {
          $job{'send_i'}="-i";
     }

     if ($job{'recursiv'}=~/yes|true|1/) {
          $job{'send_rec'}="-R";
     }  else {
         $job{'send_rec'}="";
     }

     # replace dest_ip ncsend_ip for dest 127.0.0.1
     $cmd="l_$id zfs send $job{'send_option'} $job{'send_raw'} $job{'send_rec'} $job{'send_i'} $newest_src_snap $next_src_snap | nc -w 30 $job{'ncsend_ip'} $job{'port'} &";
     $cmd=~s/ +/ /g;

     print "\n";    # nl
     $r=&socket("$cmd",$job{'source_ip'});
     if ($r=~/error/)  {  &my_log_last("start sender on $job{'source_ip'}\n$r");  &my_fast_end;  exit; }

     &my_log_last("start sender on $job{'source_ip'}","$cmd");
     &my_log_id("start sender on $job{'source_ip'}","$cmd");


     sleep(2);
     ###########################
     #verify remote sender cmd
     ###########################
     $r=&socket("rl",$job{'source_ip'});            # remote log
     if ($r=~/error/) {   &my_fast_end;  exit;  }

     @t=split(/\n/,$r);
     @t=reverse(@t);

     #  zfs send tank/data@1711788007_repli_zfs_192.168.2.75_nr_1 | /var/web-gui/cs_server/tools/nc/bsd/nc -w 30 192.168.2.76 58527 2>&1 >>/var/web-gui/cs_server/tmp/remote.log &
     foreach my $t (@t) {
         if (($t=~/$job{'port'}/) && ($t=~/zfs send/))    {
                  &my_log_id("send cmd on $job{'source_ip'}","$t");
                  print __LINE__." $job{'source_ip'}: $t\n";
                  last;
         }
     }

        ##########################
        # wait for receiver to end
        ##########################

        $t1=time();

        # get size of source fs
        my $size=&socket("zfs list -H -t filesystem,volume -o used | grep $job{'source_zfs'}",$job{'source_ip'});
        if ($size=~/error/)  {   &my_fast_end;  exit; }

        # job zfs-source size needed in monitor
        $size=&my_val2mb($size);
        $r="$size\_";
        &my_write("$wpath/_log/jobs/$id.progress",$r);

        # set timeout error as default
        &my_log_timeout;

        # ended when next_dest_snap is created + update progress
        my $dsize;         # destination size
        my $last_dsize=0;  # last destination size
        my $percent;       # percent done
        my $timeout=0;

        my $dest_parent=$job{'destination_zfs'};
        $t=rindex($job{'destination_zfs'},"/");              # parent dest fs
        $dest_parent=substr($job{'destination_zfs'},0,$t);
        print __LINE__." dest parent: $dest_parent\n";
        my ($dest_snaps,$dest_fs);

        print "\n";
        while (1) {
             sleep(40);                # check after/once every 40s

              # next dest snap created?
              $dest_snaps=&socket("zfs list -r -H -o name -t snapshot $dest_parent | grep $id",$job{'destination_ip'},120);     #timeout 120s
              if ($dest_snaps=~/error/)  {  &my_log_last("zfs list -r -H -o  name -t snapshot $dest_parent | grep $id <br>$dest_snaps"); &my_fast_end;  exit; }
              &my_log_last("dest snaps: \n$dest_snaps");

              $dest_fs=&socket("zfs list -r -H -o name -t filesystem $dest_parent",$job{'destination_ip'},120);     #timeout 120s
              if ($dest_fs=~/error/)  {   &my_fast_end;  exit; }

              if ($dest_snaps=~/$next_dest_snap/s) {

                  print __LINE__." next dest snap: $next_dest_snap is created\n";
                  &my_log_last("next dest snap created<br>$next_dest_snap");
                  &my_log_id("next dest snap created<br>$next_dest_snap");
                  $ok="1";
                  last;

              } else {

                  # update progress, source size in $size   id.progress:  sourcesize_destsize_percent
                  # get size of dest fs

                  if ($dest_fs=~/$job{'destination_zfs'}/s) {
                        #
                        $dsize=&socket("zfs list -H -t filesystem,volume -o used $job{'destination_zfs'}",$job{'destination_ip'});
                        if ($dsize=~/error/)  {  &my_fast_end;  exit; }

                        $dsize=&my_val2mb($dsize);

                        if ($size > 0) {          # source size

                           $percent=int($dsize/$size*100);  # 3.1 -> 3%
                           &my_write("$wpath/_log/jobs/$id.progress","$size\_$dsize\_$percent");


                           if ($dsize > $last_dsize) {

                                 $timeout=0;         # reset timeout
                                 &my_log_last("wait until next dest snap is created: $next_dest_snap, $percent % done");
                           }  else {
                                 print __LINE__." wait until next dest snap is created: $next_dest_snap, $percent % done\n";
                           }

                        }

                    }  else {

                         print __LINE__." wait for creation of dest fs $job{'destination_zfs'}\n";

                         $r=&socket("ps axw | grep zfs",$job{'destination_ip'});
                         if ($r=~/error/)  { &my_fast_end;  exit; }

                         $t=&socket("rl | grep receiving",$job{'destination_ip'});
                         if ($t=~/error/)  { &my_fast_end;  exit; }
                         $t=~s/.*receiving/receiving/s;     # only last

                         &my_log_last("processes on destination $job{'destination_ip'}\n$r\n$t");
                    }
              }

              $last_dsize=$dsize;
              $timeout++;       # ??
              if ($timeout > 2) {
                    print __LINE__." receiver timeout\n";
                    &my_log_last("receiver timeout");  &my_fast_end;  exit;
              }
        }

        #################### /end waiting


     ##########################
     # postprocess keep, hold?
     #########################

     #read new snaps
     $t=&socket("zfs list -H -r -o name -t snapshot -s creation",$job{'destination_ip'});       # newest last
     if ($t=~/error/)  {  &my_fast_end;  exit; }

     @s=split(/\n/,$t);
     @s=reverse(@s);  # newest first to hold last n newest snaps

     $t=$id."_recursiv";
     $t=$current{$t};                                             # yes

     if ($t=~/1|y/i) {
         @s=sort(@s);                 # sort name to allow keep/hold recursive
         if ($debug=~/3/) { print "\nxxxxxxxxx\nsort name\n"; }
     }


     if ($ok) {                    # next dest snap created
          my $dnr;
          my $dnr2;
          my $dnr3;
          if ($newest_dest_snap=~/_nr_\d+/) {
                   $dnr=$newest_dest_snap;
                   $dnr=~s/.*_nr_//;                                      # current
                   $dnr2=$dnr+1;                                          # next snapnr
                   $dnr3=$dnr-1;                                          # previous

          }
          if (($newest_src_snap ne "")  && ($id>0) && ($dnr>1) && ($newest_dest_snap ne "" ))     {

                     # process local source snaps
                     foreach my $s (@s) {
                         if (!($s=~/$id\_repli/))        { next; }        # not my snaps
                         if ($s=~/_nr_$dnr2$/)           { next; }        # next nr   (keeps 2 snaps, comment for one snap)
                         if ($s=~/_nr_$dnr$/)            { next; }        # current  nr
                         if ($s=~/_nr_$dnr3$/)           { next; }        # previous  nr

                         if ($s=~/^$job{'destination'}/) { next; }        # this is a destination snap

                         #keep not on source
                          if ($s eq $newest_dest_snap)   { next; }
                          $t=&socket("zfs destroy -f $s",$job{'source_ip'});
                          if ($t=~/error/)  {  &my_fast_end;  exit; }
                     }

                     # process local dest snaps
                     foreach my $s (@s) {
                         if (!($s=~/$id\_repli/))        { next; }        # not my snaps
                         if ($s=~/_nr_$dnr2$/)           { next; }        # next nr   (keeps 2 snaps, comment for one snap)
                         if ($s=~/_nr_$dnr$/)            { next; }        # current  nr
                         if ($s=~/_nr_$dnr3$/)           { next; }        # previous  nr
                         if (!($s=~/^$job{'destination'}/)) { next; }     # this is a source snap


                         #keep
                          if (&my_keep($s,$dnr))         { next; }        # process keep on "destination fs"
                          if ($s eq $newest_dest_snap)   { next; }

                          $t=&socket("zfs destroy -f $s",$job{'source_ip'});
                          if ($t=~/error/)  {  &my_fast_end;  exit; }
                     }



         }
         ###################
         ## end with success
         ###################

         &my_ok_and_exit;

     }

     ############# not ok ########################
     &my_fast_end;

}

 ################################
 sub my_local_initial_replication {
 ################################

   my (@t,$t);

     &my_log_last("start local initial replication");
     my $p;
     &update_lastrun;

     &my_log_last("start local initial replication");
     &my_log_id("start local initial replication");

     # opt delete old snap with this id (restart)
     &my_del_snap($id);

     #create first source snap, opt recursiv
     my $snap="$job{'source_zfs'}\@$id\_repli_zfs_127.0.0.1_nr_1";
     my $cmd="l_$id zfs snapshot $job{'snap_rec'} $snap"; $cmd =~s/ +/ /g;

     &my_log_id("cmd","$cmd");
     &my_log_last("cmd: $cmd");

     my $r=&socket("$cmd",$job{'source_ip'});           # snap
     if ($r =~/error/)  {  &my_fast_end;  exit; }

     &log("$id <br>job-replicate 1302  | create snap | $cmd | $r ","m");                   # monitor

     my $next_dest_snap="$job{'destination_zfs'}\@$id\_repli_zfs_127.0.0.1_nr_1";         # backup3/nfs.www@1330674968_repli_zfs_backup3_nr_266

     #debug
     # print "job-replication 741: local init size :$job{'zfs_source_size'}\n$cmd\n$r";

     if ($job{'send_raw'}=~/yes|true|1/) {
          $job{'send_raw'}="--raw";
     }  else {
              $job{'send_raw'}="";
     }

     if ($job{'recursiv'}=~/yes|true|1/) {
               $job{'send_rec'}="-R";
     }  else {
              $job{'send_rec'}="";
     }


     my $send="zfs send $job{'send_rec'} $job{'send_raw'} $job{'send_options'}  $snap"; $send=~s/ +/ /g;
     my $rcv="zfs receive $job{'rec_options'} -Fv $job{'destination_zfs'}";

     $cmd="l_$id $send | $rcv &";     $cmd =~s/ +/ /g;

     &my_log_id("cmd","$cmd");
     &my_log_last("$cmd");

     $current{'receiver_pids'}=&socket("$cmd",$job{'source_ip'});                 # background
     if ($r =~/error/) {  &my_fast_end;  exit; }

     # send ok

        ##########################
        # wait for receiver to end
        ##########################

        $t1=time();

        # get size of source fs
        my $size=&socket("zfs list -H -t filesystem,volume -o used $job{'source_zfs'}",$job{'source_ip'});
        if ($size=~/error/)  {  &my_fast_end;  exit; }

        # job zfs-source size needed in monitor
        $size=&my_val2mb($size);
        $r="$size\_";
        &my_write("$wpath/_log/jobs/$id.progress",$r);

        # set timeout error as default
        &my_log_timeout;

        # ended when next_dest_snap is created + update progress
        my $next_dest_snap ="$job{'destination_zfs'}\@$id\_repli_zfs_127.0.0.1\_nr_1";           # local transfer
        my $dsize;         # destination size
        my $last_dsize=0;  # last destination size
        my $percent;       # percent done
        my $timeout=0;

        my $dest_parent=$job{'destination_zfs'};
        $t=rindex($job{'destination_zfs'},"/");              # parent dest fs
        $dest_parent=substr($job{'destination_zfs'},0,$t);
        print "751 dest parent: $dest_parent\n";
        my ($dest_snaps,$dest_fs);

        print "\n";
        while (1) {
             sleep(40);                # check after/once every 40s

              # next dest snap created?
              $dest_snaps=&socket("zfs list -r -H -o name -t snapshot $dest_parent | grep $id",$job{'destination_ip'},120);     #timeout 120s
              if ($dest_snaps=~/error/)  {  &my_fast_end;  exit; }
              print "1374\n$dest_snaps\n";
              $dest_fs=&socket("zfs list -r -H -o name -t filesystem $dest_parent",$job{'destination_ip'},120);     #timeout 120s
              if ($dest_sf=~/error/)  {  &my_fast_end;  exit; }

              if ($dest_snaps=~/$next_dest_snap/s) {

                  print __LINE__."next dest snap: $next_dest_snap is created\n";
                  last;

              } else {

                  # update progress, source size in $size   id.progress:  sourcesize_destsize_percent
                  # get size of dest fs

                  if ($dest_fs=~/$job{'destination_zfs'}/s) {
                        #
                        $dsize=&socket("zfs list -H -t filesystem,volume -o used $job{'destination_zfs'}",$job{'destination_ip'});
                        if ($dsize=~/error/)  {   &my_fast_end;  exit; }

                        $dsize=&my_val2mb($dsize);

                        if ($size > 0) {

                           $percent=int($dsize/$size*100);  # 3.1 -> 3%
                           &my_write("$wpath/_log/jobs/$id.progress","$size\_$dsize\_$percent");

                           print __LINE__."wait until next dest snap is created: $next_dest_snap, $percent % done\n";
                           if ($dsize > $last_dsize) {  $timeout=0; }              # reset timeout
                        }
                    }  else {
                         print __LINE__." wait for creation of dest fs $job{'destination_zfs'}\n";
                    }
              }
              $timeout++;
              if ($timeout > 3) {
                    print __LINE__."receiver timeout\n";
                    &my_log_id("receiver timeout"); $error="error connect";  &my_fast_end;  exit;
              }
        }

        #################### /end waiting

        ###################
        ## end with success
        ###################

        &my_ok_and_exit;

 }


 ################################
 sub my_local_delta_replication {
 ################################

     &my_log_last("1436 next local incremental replication");

     my $p;
     &update_lastrun;

     &my_log_last("start local incremental replication");
     &my_log_id("start local incremental replication");

     #step 1 check for newest source dest snaps
     my $newest_src="0";
     my $newest_dest="0";

     my $newest_src_snap="";
     my $newest_dest_snap="";
     my $next_src_snap="";
     my $next_dest_snap="";
     my $last_dest_snap="";

     ($newest_src_snap,$newest_dest_snap,$next_src_snap,$next_dest_snap) =&my_newest_snap_pair();  # local transfer, no par

     #debug
     #print "\n ## local delta 819 \nns $newest_src_snap\nnd $newest_dest_snap\nxs $next_src_snap\nxd $next_dest_snap\n";
     #exit;

     # get local snaps and opt delete next src snap (kept from unsuccessfull repli)
     my $t=&socket("zfs list -H -r -o name -t snapshot -s creation $job{'source_zfs'}",$job{'source_ip'});  # newest last dest
     if ($t=~/error/)  {  &my_fast_end;  exit; }
#&mess($t);

     my @s=split(/\n/,$t);

     # opt delete next (from failed previous)
     if (($next_src_snap=~/$id\_repli/) && ($id>0))  {
        foreach my $s (@s) {
          if ($s eq "$next_src_snap") {
             $t=&socket("zfs destroy -f $s",$job{'source_ip'});
             if ($t=~/error/)  {  &my_fast_end;  exit; }
          }
        }
     }

     #create next source snap
     my $cmd="l_$id zfs snapshot $job{'snap_rec'} $next_src_snap"; $cmd =~s/ +/ /g;
     my $r=&socket("$cmd",$job{'destination_ip'});
     if ($r=~/error/) {  &my_log_last("create next src snap: $r"); &my_fast_end;  exit; }

     &log("$id <br>job-replicate 1479 | create snap | $cmd | $r ","m");     #monitor
     #&my_log_id("result","$r");    # keep error
     &my_log_last("result\n$r");

     if ($job{'send_i'} ne "-I") {
          $job{'send_i'}="-i";
     }

     #prepare transfer command
     if ($job{'send_raw'}=~/yes|true|1/) {
          $job{'send_raw'}="--raw";
     }  else {
              $job{'send_raw'}="";
     }

     if ($job{'send_i'} ne "-I") {
               $job{'send_i'}="-i";
     }

     if ($job{'recursiv'}=~/yes|true|1/) {
               $job{'send_rec'}="-R";
     }  else {
              $job{'send_rec'}="";
     }

     my $send="zfs send $job{'send_i'} $job{'send_rec'} $job{'send_raw'} $job{'send_option'}  $job{'send_rec'} $newest_src_snap $next_src_snap";  $send=~s/ +/ /g;
     my $rcv="zfs receive $job{'rec_option'} -Fv $job{'destination_zfs'}";                                                   $rcv =~s/ +/ /g;
     my $cmd="l_$id $send | $rcv &";

     &my_log_id("cmd","$cmd");
     &my_log_last("cmd\n","$cmd");

     # start cmd
     $current{'receiver_pids'}=&socket("$cmd",$job{'source_ip'});                 # background
     &my_log_last("start replication: $current{'receiver_pids'}");
     if ($current{'receiver_pids'} =~/error/)  { &my_log_last("start cmd: $current{'receiver_pids'}");  &my_fast_end;  exit; }

     # send ok

        ##########################
        # wait for receiver to end
        ##########################

        $t1=time();

        # get size of source fs
        my $size=&socket("zfs list -H -t filesystem,volume -o used $job{'source_zfs'}",$job{'source_ip'});
        if ($size=~/error/)  { &my_log_last("size: $size");  &my_fast_end;  exit; }

        # job zfs-source size needed in monitor
        $size=&my_val2mb($size);
        $r="$size\_";
        &my_write("$wpath/_log/jobs/$id.progress",$r);

        # set timeout error as default
        &my_log_timeout;

        # ended when next_dest_snap is created + update progress
        my $dsize;         # destination size
        my $last_dsize=0;  # last destination size
        my $percent;       # percent done
        my $timeout=0;

        my $dest_parent=$job{'destination_zfs'};
        $t=rindex($job{'destination_zfs'},"/");              # parent dest fs
        $dest_parent=substr($job{'destination_zfs'},0,$t);
        print __LINE__." dest parent: $dest_parent\n";
        my ($dest_snaps,$dest_fs);

        print "\n";

        while (1) {

              sleep(40);
              # next dest snap created?
              $dest_snaps=&socket("zfs list -r -H -o name -t snapshot $dest_parent | grep $id",$job{'destination_ip'},120);     #timeout 120s
              if ($dest_snaps=~/error/)  { &my_log_last("list snaps: $sdest_snaps");   &my_fast_end;  exit; }
              print "1548\n$dest_snaps\n";

              $dest_fs=&socket("zfs list -r -H -o name -t filesystem $dest_parent",$job{'destination_ip'},120);     #timeout 120s
              if ($dest_fs=~/error/)  { &my_log_last("list fs: $dest_fs");  &my_fast_end;  exit; }

              if ($dest_snaps=~/$next_dest_snap/s) {

                  &my_log_last("next dest snap created<br>$next_dest_snap");
                  &my_log_id("next dest snap created<br>$next_dest_snap");
                  last;

              } else {

                  # update progress, source size in $size   id.progress:  sourcesize_destsize_percent
                  # get size of dest fs

                  if ($dest_fs=~/$job{'destination_zfs'}/s) {
                        #
                        $dsize=&socket("zfs list -H -t filesystem,volume -o used $job{'destination_zfs'}",$job{'destination_ip'});
                        if ($dsize=~/error/)  {  &my_log_last("dest size: $dsize");  &my_fast_end;  exit; }

                        $dsize=&my_val2mb($dsize);

                        if ($size > 0) {

                           $percent=int($dsize/$size*100);  # 3.1 -> 3
                           &my_write("$wpath/_log/jobs/$id.progress","$size\_$dsize\_$percent");

                           print __LINE__." wait until next dest snap is created: $next_dest_snap, $percent % done ($size\_$dsize\_$percent)\n";
                           if ($dsize > $last_dsize) {  $timeout=0; }              # reset timeout

                        }
                    }  else {
                         print "1580 wait for creation of dest fs $job{'destination_zfs'}\n";
                    }
              }

              $timeout++;
              if ($timeout > 3) {
                    print "1586 timeout\n";
                    &my_log_id("error 1587, receiver timeout 2min"); $error="error connect";  &my_fast_end;  exit;
              }
        }

      #################### /end waiting

     ##########################
     # postprocess keep, hold?
     #########################


     #read new dest snaps
     $t=&socket("zfs list -H -r -o name -t snapshot -s creation | grep $job{'destination_zfs'}",$job{'destination_ip'});       # newest last
     if ($t=~/error/)  {  &my_log_last("list snap: $t");  &my_fast_end;  exit; }



     &my_log_last("new dest snaps:\n$t");
     @s=split(/\n/,$t);
     @s=reverse(@s);  # newest first to hold last n newest snaps

     $t=$id."_recursiv";
     $t=$current{$t};                                             # yes

     if ($t=~/1|y/i) {
         @s=sort(@s);                 # sort name to allow keep/hold recursive
         if ($debug=~/3/) { print "\nxxxxxxxxx\nsort name\n"; }
     }

     my $ok="0";
     foreach my $s (@s) {
          # print "job-repli 1615 snap $s\n";
          if ($s eq "$next_dest_snap") {             #ok
              $ok="1";
          }
     }

     # delete all but newest two pairs
     if ($ok) {
          my $dnr;
          my $dnr2;
          my $dnr3;
          if ($newest_dest_snap=~/_nr_\d+/) {
                   $dnr=$newest_dest_snap;
                   $dnr=~s/.*_nr_//;                                      # current
                   $dnr2=$dnr+1;                                          # next snapnr
                   $dnr3=$dnr-1;                                          # previous

          }
          if (($newest_src_snap ne "")  && ($id>0) && ($dnr>1) && ($newest_dest_snap ne "" ))     {

                     # process local source snaps
                     foreach my $s (@s) {
                         if (!($s=~/$id\_repli/))        { next; }        # not my snaps
                         if ($s=~/_nr_$dnr2$/)           { next; }        # next nr   (keeps 2 snaps, comment for one snap)
                         if ($s=~/_nr_$dnr$/)            { next; }        # current  nr
                         if ($s=~/_nr_$dnr3$/)           { next; }        # previous  nr

                         if ($s=~/^$job{'destination'}/) { next; }        # this is a destination snap

                         #keep not on source
                          if ($s eq $newest_dest_snap)   { next; }
                          $t=&socket("zfs destroy -f $s",$job{'source_ip'});
                          if ($t=~/error/)  {  &my_fast_end;  exit; }
                     }

                     # process local dest snaps
                     foreach my $s (@s) {
                         if (!($s=~/$id\_repli/))        { next; }        # not my snaps
                         if ($s=~/_nr_$dnr2$/)           { next; }        # next nr   (keeps 2 snaps, comment for one snap)
                         if ($s=~/_nr_$dnr$/)            { next; }        # current  nr
                         if ($s=~/_nr_$dnr3$/)           { next; }        # previous  nr
                         if (!($s=~/^$job{'destination'}/)) { next; }     # this is a source snap


                         #keep
                          if (&my_keep($s,$dnr))         { next; }        # process keep on "destination fs"
                          if ($s eq $newest_dest_snap)   { next; }

                          $t=&socket("zfs destroy -f $s",$job{'source_ip'});
                          if ($t=~/error/)  {  &my_log_last("snap del: $t");  &my_fast_end;  exit; }
                     }
         }

        ###################
        ## end with success
        ###################

        &my_ok_and_exit;
        ################

     }

     #error:

     &my_log_last("next destination snap $next_dest_snap was not created");
     &my_fast_end;
     exit;


 }

 ###################
 sub my_ok_and_exit {
 ###################

       my @p1=caller(0); # this:
       my @p2=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask
       # calling line: $p1[2]

       print "line $p1[2] ok and exit\n";

       ###################
        ## end with success
        ###################

        &my_postreplication;

        # log dest result #
        ##################
        my $t2=time();
        sleep(2);
        $t=$t2-$t1;
        $r=&socket("rl",$job{'destination_ip'});
        if ($r=~/error/)  { &my_log_last("remote log: $r");  &my_fast_end;  exit; }

        #zfs send tank/data@1711876166_repli_zfs_127.0.0.1_nr_1 | zfs receive -Fv tank/data4 2>&1 >>/xampp/web-gui/data/cs_server/tmp/remote.log &
        #receiving full stream of tank/data@1711876166_repli_zfs_127.0.0.1_nr_1 into tank/data4@1711876166_repli_zfs_127.0.0.1_nr_1
        #received 47.9K stream in 0.16 seconds (305K/sec)

        $r=~s/.*receiving.+\@$id//s; # last entry with jobid
        if ($r=~/received/s) {
             $r=~s/.*received/received/s;   $r=~s/\n.*//s;  chomp($r);
        } else {
             $r="remote log info: n.a.";
        }

        print __LINE__."\n ok: $r\n";
        # /log result

        &my_log_id("ok: replication run ends after $t s","$r");
        &my_log_last("ok: replication run ends after $t s","$r");
        &my_log_jobs("ok: replication $id run ends after $t s");

        if (-f "$wpath/_log/jobs/$id.progress")     {  unlink("$wpath/_log/jobs/$id.progress"); }  # progress
        if (-f "$wpath/_log/jobs/$id.running")      {  unlink("$wpath/_log/jobs/$id.running"); }   #
        if (-f "$wpath/_log/jobs/$id.sender")       {  unlink("$wpath/_log/jobs/$id.sender"); }   #


        ################################
        exit;
        ################################


 }

#######################     get newest source/dest snap-pair  from local replication
sub my_newest_snap_pair {
#######################     parameter: source snap array on remote transfers

  print "\n";              # cli nl

  my @src_snaps=();
  my @dst_snaps=();

  # src snaps
  my $t=&socket("zfs list -H -r -o name -t snapshot -s creation | grep repli,$id,$job{'source_zfs'}\@",$job{'source_ip'});            #  all source repli snaps for id and repli oldest first
  if ($t=~/error/s)  {  &my_log_last("list src snaps: $t");  &my_fast_end;  exit; }
  @src_snaps=split(/\n/,$t);
  &my_log_last("list of repli snaps on $job{'source_ip'} for id $id and $job{'source_zfs'}\@: (src)\n$t");

  print "\n";               # cli nl
  # dst snaps
  $t=&socket("zfs list -H -r -o name -t snapshot -s creation | grep repli,$id,$job{'destination_zfs'}\@",$job{'destination_ip'});  #  all dest repli snaps for id and fs
  if ($t=~/error/s)  {  &my_log_last("list dst snap: $t");  &my_fast_end;  exit; }

  @dst_snaps=split(/\n/,$t);
  &my_log_last("list of repli snaps on $job{'destination_ip'} for id $id and $job{'destination_zfs'}\@: (dst)\n$t");
  print "\n";               # cli nl
                            # @snaps: local and remote
  my $newest_src_nr="0";
  my $newest_dest_nr="0";

  my $newest_src_snap="";   # newest common=base snap
  my $next_src_snap="";
  my $next_dest_snap="";
  my $newest_dest_snap="";               # newest common=base snap

  # get newest destination (last successfull replication)
  # print "job-repli 946: search $job{'destination'}\@$id\_repli\n";
  # ex backup3/157_daten@1379949048_repli_zfs_rz-backup3_nr_1

  my $ok1="0";
  my ($sref,$dref,$td);

  foreach my $ds (@dst_snaps) {                 # destination snaps (local or remote) sorted by date, oldest first

          if (!($ds=~/^$job{'destination_zfs'}\@$id\_repli/)) { next; }  # receiver/my snaps: backup3/nfs.www@1330674968_repli_zfs_backup3_nr_266      (destination - > myhost)
          if (!($ds=~/\_nr\_\d+$/)) { next; }                            # valid repli nr

          # this must be our oldest or a newer destination snap
          $td=$ds;                       # maybe newest_dest_snap

          # now check if there is a correspondent src snap (This is indeed a valid base snap then)
          # condition: same string up from @
          # src:       tank/data@1712000135_repli_zfs_192.168.2.75_nr_1
          # dst: daten1/test/bsd@1712000135_repli_zfs_192.168.2.75_nr_1

          $dref=$ds;
          $dref=~s/.*\@/\@/;                    # up from @
print "dref=$dref\n";
          $ok1="0";
          foreach my $ss (@src_snaps) {        # sorted by creation date
                  $sref=$ss;
                  $sref=~s/.*\@/\@/;           # up from @
print "sref=$sref\n";
                  if ($dref eq $sref)  {       # we have a match for newest base snap
                      $newest_src_snap=$ss;
                      $t=$ss; $t=~s/.*_nr_//;
                      $newest_src_nr=$t;

                      $newest_dest_snap=$td;
                      $t=$td; $t=~s/.*_nr_//;
                      $newest_dest_nr=$t;
                      last;
                  }
          }
  }


  if (($sref eq $dref) && ($sref ne ""))  {

       &my_log_last("source-dest snap-pair $sref found");

       #  next nr:
       my $next=$newest_dest_nr+1;
       $next_src_snap=$newest_src_snap;
       $next_src_snap  =~s/\_$newest_dest_nr$/_$next/;            # _nr_2 -> _nr_3

       $next_dest_snap=$newest_dest_snap;
       $next_dest_snap =~s/\_$newest_dest_nr$/_$next/;

  } else {                                      # cancel; error

           &my_log_last("source-dest snap-pair $newest_dest_nr not found.");
           &my_log_id("source-dest snap-pair $newest_dest_nr not found, check repli_snaps on source and target");   $info=~s/ +/ /g;

           &my_fast_end;
           exit;
  }


  #return ($newest_src_snap, $next_src_snap)
  $job{'next_src_snap'}=$next_src_snap;
  print __LINE__." snappair newest_src: $newest_src_snap\n";
  print __LINE__." snappair newest_dest: $newest_dest_snap\n";
  print __LINE__." snappair next_src: $next_src_snap\n";
  print __LINE__." snappair next_dest: $next_dest_snap\n";

  return($newest_src_snap,$newest_dest_snap,$next_src_snap,$next_dest_snap);
}

############
sub my_keep {
############
   # keep and hold snaps on dest
   # &my_keep($s,$dnr): snap + newest dest snapnr

   #######################################################
   # keep and hold does not work with rec replication !!!!
   #######################################################

   my $min;
   my $s=$_[0];    # process snaps with newest=last
   my $curr=$_[1]; # current number

   if ($s eq $current{'newest_dest_snap'}) {  return('1'); }                # keep always
   if ($s eq $current{'last_dest_snap'})   {  return('1'); }

   #get filesystem
   $t=$s;
   $t=~s/\@.*//s;

   if ($t ne $keepfs) {                   # first or new fs, to allow recursive replications
        $keepfs=$t;
        @keeph=();
        @keepd=();
        @keepm=();
        @keepy=();
        $holdnr="0";
   }


   my $t=$id."_keep";                     # keep is a commalist like hours:24,days:32,months:12,years:2,.00 (last is char)
   $t=$current{$t};
   $t=~s/ +//g;

   my $hold=$id."_hold";                  # keep is a commalist like hours:24,days:32,months:12,years:2,.00 (last is char)
   $hold=$current{$hold};

   if ($debug=~/3/) { print "\n1469 sub my_keep process keep/hold snap:\n  $s\n"; print "  keep $t, hold $hold\n"; }

   my @keep=split(/,/,$t);
   my (@char,$char,$hours,$days,$months,$years,$n,@w,$r,$snapdate,$lastnr);


   $snapdate=&socket("zfs list -H -o creation $s",$job{'destination'});  # ex Sun Sep 15  0:05 2013
   if ($t=~/error/)  { &my_log_last("list creation: $t");  &my_fast_end;  exit; }

   foreach my $t (@keep) {
      if ($t eq "") { next; }
      if ($t=~/hour.*\d/) {
        $t=~s/\D//g;
        if ($t == 0) { $t="1"; }               # hour 1..24
        $hours=$t;
        next;
      }
      if ($t=~/day.*\d/)   { $t=~s/\D//g; $days=$t;   next; }
      if ($t=~/month.*\d/) { $t=~s/\D//g; $months=$t; next; }
      if ($t=~/year.*\d/)  { $t=~s/\D//g; $years=$t;  next; }


      #char
      push (@char,$t);
   }

   # 1. check keep char with rename (keep outside retention)
   ##################
   if (@char>0) {
      foreach my $p (@char) {
        if ($p eq "")  { next; }
        if ($s=~/$p$/) {
             $n=$s;
             $n=~s/.*\_//;
             &my_log_id("1499 keep nr $n","keep $t: match rule: $p");
             #rename to keep it even when rule changes

             $t=$s; $t=~s/repli/keep/;
             $t=~s/\_nr/\_was/;                # or autodelete will happen on _nr_n
              $t=&socket("zfs rename $s $t",$job{'destination'});
              if ($t=~/error/) {  &my_log_last("rename: $t");  &my_fast_end;  exit; }

              if ($debug=~/3/) { print "1505 keep char >>>>> 1505 rename $s\n$t\n\n"; }
             return('1');                      # skip delete
        }
      }
  }

  @w=();
  if (($hours>0) || ($days>0) || ($months>0) || ($years>0)) {
     $r=$snapdate;
     chomp($r);
     $r=~s/ +/ /g;

     @w=(); @w=split(/ +/,$r);

     #month in w[1], day in w[2], year in w[4]
     # month can be mai Mai May (depends on localisation)
     $w[1]=~s/Jan/01/i;     $w[1]=~s/Feb/02/i; $w[1]=~s/Mar/03/i; $w[1]=~s/Apr/04/i;
     $w[1]=~s/Ma[ijy]/05/i; $w[1]=~s/Jun/06/i; $w[1]=~s/Jul/07/i; $w[1]=~s/Aug/08/i;
     $w[1]=~s/Sep/09/i;     $w[1]=~s/^O.*/10/i; $w[1]=~s/Nov/11/i; $w[1]=~s/De.*/12/i;

     if ($w[2] < 10) {         # 3 -> 03
       $w[2]="0$w[2]";
       $w[2]=~s/^0+/0/;
     }

     $w[3]=~s/:.*//;    # hour 0..23 only
     if ($debug=~/3/) { print "  1527 snapdate $snapdate from zfs get creation: $r\n    day dd.mm.yyyy hour = $w[0] $w[2].$w[1].$w[4] hour: $w[3] vs today $sys{'mday'}.$sys{'mon'}.$sys{'year'} hour $sys{'hour'}\n"; }
  }


  # 2. keep on snap per last 1..n years
  if ($years > 0) {
      # how old is snap in years
      $t=0;    #
      for (my $y=$sys{'year'};$y>=$w[4];$y--) {           # this=1
              $t++;
      }

      if ($debug=~/3|y/) { print "  1539 process years=$years with snap-age in years (this=1): $t\n"; }

      # keep year
      if (($keepy[$t] ne "1") && ($t<=$years)) {            # global array of years, this year keepy[1]
             $keepy[$t]="1";              # remember

             $t=$s;
             $t=~s/_zfs\-..\d+_/_zfs_/;   # opt remove old rules, old script had used hn30s for 30 snaps!
             $t=~s/_zfs_/_zfs-ky$w[4]\_/;
             if ($s ne $t) {
                 `$ZFS rename $s $t`;
                 if ($debug=~/3|y/) { print "      1550 rename and keep year  $w[4] $s\n        to $t\n"; }        # keep the oldest from this year
             }  else {
                 if ($debug=~/3|y/) { print "      1553 keep year $w[4] $s\n"; }                         # keep the oldest from this year
             }
             return('1');       # skip delete
      }  else {
             if ($debug=~/3|y/) { print "    1558 year $w[4] is already kept or > maxmyears\n"; }
      }
  }



  # 3. keep on snap per last 1..n months     (debug 3c)
  if ($months > 0) {
     #how old is snap in months
#$w[4]--; # debug

     if ($debug=~/[3789m]/) { print "\n1756 process $s\n months=$months on $w[1] $w[4] \n"; }


     $t=0;    #
     for (my $y=$w[4];$y<=$sys{'year'};$y++) {                   # snapyear to this year

        if (($y==$w[4]) && ($y != $sys{'year'})) {               # old year of snap
          for (my $m=$w[1];$m<12;$m++) {                         # til end of year
            $t++;
          }
          if ($debug=~/7/) { print " 1762 snapyear to end year $w[4]: $t\n"; }
          next;
        }

        if (($y>$w[4]) && ($y<$sys{'year'})) {                   # years between
          $t=$t+12;
          if ($debug=~/8/) { print " 1771 years between: $t\n"; }
          next;
        }

        # snap is from older years
        if ($w[4] != $sys{'year'}) {
          for (my $m=1;$m<=$sys{'mon'};$m++) {                   # til this month and old year
            $t++;
          }
          if ($debug=~/9/) { print " 1777 this year to this month $sys{'mon'}: $t\n"; }

        }  else {

          # snap is from this year

          for (my $m=$w[1];$m<$sys{'mon'};$m++) {               # til this month and snap this year
            $t++;
          }
          if ($debug=~/9/) { print " 1788 this year to this month $sys{'mon'}: $t\n"; }
        }
      }

      if ($debug=~/[3789]|m/) { print " 1783: month-diff=$t\n\n"; }
      $t++; # diff=0 -> $t=1

      # keep month
      if (($keepm[$t] eq "") && ($t<=$months)) {
             $keepm[$t]="1";    # remember

             $t=$s;
             $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
             my $yr=substr($w[4],2,2);
             $t=~s/_zfs_/_zfs-km$yr$w[1]\_/;
             if ($s ne $t) {
               $r=&socket("zfs rename $s $t",$job{'destination'});
               if ($r=~/error/)  {  &my_log_last("rename: $r");  &my_fast_end;  exit; }

               if ($debug=~/3|m/)  { print "    1974 rename and keep month $w[1] $w[4] $s\n$t\n\n"; }
             } else {
               if ($debug=~/3|m/)  { print "    1976 keep month $w[1] $w[4] $s\n"; }                         # keep the oldest from this year
             }

             return('1');       # skip delete
      } else {
             if ($debug=~/3|m/) { print "    1590 month $w[1] $w[4] (month age=$t) is already kept or > max months\n"; }
      }
  }

  # ex Sun Sep 15  0:05 2013
  # 4. keep one snap per last 1..n days
  if ($debug=~/3|d/) { print "  1688 process $s with $snapdate and days $days with date=$w[2].$w[1].$w[4]\n"; }
  if ($days > 0) {
      $t=&date2diff("$w[2].$w[1].$w[4]","$sys{'mday'}.$sys{'mon'}.$sys{'year'}");
      if ($debug=~/3|d/) { print "    1553 process days $days\: $w[2].$w[1].$w[4] >> $sys{'mday'}.$sys{'mon'}.$sys{'year'}\n 1553 diff=$t\n"; }

      if ($t <= $days)   {
         # keep day

         $r="$w[1]$w[2]";
         if ($keepd[$r] eq "") {
             $keepd[$r]="1";            # remember
             if ($debug=~/3|d/) { print "    1524: process day: $w[2] with diff=$t\n"; }

             $t=$s;
             $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
             $t=~s/_zfs_/_zfs-kd$w[1]$w[2]\_/;
             if ($s ne $t) {
                `$ZFS rename $s $t`;
                if ($debug=~/3|d/)  { print "      1615 rename and keeo day $w[2] $s\n$t\n"; }
             }  else {
                 if ($debug=~/3|d/) { print "      1617 keep day $w[2] $s\n"; }                         # keep the oldest from this year
             }

             return('1');        # skip delete
         }
      }  else {
#debug
#print "!\n"; exit;
             if ($debug=~/3|d/) { print "    1558 day $w[2] ($t) is already kept\n"; }
      }
  }

  # 5. keep last n snaps per hour
  if ($hours > 0) {

      my $h1=&date2dop("$sys{'mday'}.$sys{'mon'}.$sys{'year'}") * 24;                # passed number of hours now
      my $h2=&date2dop("$w[2].$w[1].$w[4]") * 24;                                    # passed number of snap hours

      $h1=$h1+$sys{'hour'};    # hour value of now
      $h2=$h2+$w[3];           # hour value of snap

      $r=$h1-$h2;              # hour diff

      if ($debug=~/3|s/) { print "    1619 process hours $hours with diff=$r\n"; }
      if ($r <= $hours) {

              # @keeph: pos==data
              if ($keeph[$r] eq "") {
                   $keeph[$r]="1";    # remember
                   if ($debug=~/3|s/) { print "    1626: process hour: $r\n"; }

                   $t=$s;
                   $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
                   $t=~s/_zfs_/_zfs-kh$r\_/;
                   if ($s ne $t) {
                     `$ZFS rename $s $t`;
                     if ($debug=~/3|s/) { print "    1643 rename and keep hour $r $s\n  to $t\n"; }
                   }  else {
                     if ($debug=~/3|s/) { print "    1553 keep hour $r  $s\n"; }                         # keep the oldest from this year
                   }

                   return('1');        # skip delete
              }
      }  else {
             if ($debug=~/3|s/) { print "    1654 diff hour $r  > $hours \n"; }
      }

      #debug
      #return(1);
  }

  #######################################
  # 6. check hold (days ex 8 or snaps ex 8s)
  #######################################
  $t=$id."_hold";
  $t=$current{$t};
  my $hld=$t;
  $hld=~s/\D//g;

  if ($debug=~/3|h/) { print "  1664 process hold $t\n"; }

  # number of snaps ex 8s to hold additionally
  if ($t=~/s/i) {                     # number of snaps
                                      # use _nr as reference
       $n=$s;
       $n=~s/.*nr\_//;                # current snap nr

       if (($hld > 0) && ($n>0) && ($holdnr<=$hld))  {
          $lastnr=$current{'last_dest_snap'};
          $lastnr=~s/.*nr\_//;
          $lastnr=$lastnr-1;                                # =newest

          if ($n >= ($lastnr-$hld) ) {       # more snaps than max to hold
                  if ($debug=~/3f/) { print "1673 process hold number: $n -> $hld (last: $lastnr)\n"; }

                  $holdnr++;
                  $t=$s;
                  $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
                  $t=~s/_zfs_/_zfs-hn$holdnr\_/;
                  if ($s ne $t) {
                     $r=&socket("zfs rename $s $t",$job{'destination'});
                     if ($r=~/error/) {  &my_log_last("rename: $r");  &my_fast_end;  exit; }

                     if ($debug=~/3|h/) { print "      2090 hold $holdnr snap and rename $s\n      to $t\n"; }
                  } else {
                     if ($debug=~/3|h/) { print "      2092 hold $holdnr snap $s\n"; }                         # keep the oldest from this year
                  }
                  return('1');        # skip delete
            }  else {
                  if ($debug=~/3|h/) { print "    2096 max snaps $hld already hold\n"; }
            }

       }

  } else {

       if ($hld > 0) {

           $r=$snapdate;
           chomp($r);
           $r=~s/ +/ /g;

           @w=();
           @w=split(/ /,$r);

           #month in w[1], day in w[2], year in w[4]
           # month can be mai Mai May (depends on localisation)
           $w[1]=~s/Jan/01/i;     $w[1]=~s/Feb/02/i; $w[1]=~s/Mar/03/i; $w[1]=~s/Apr/04/i;
           $w[1]=~s/Ma[ijy]/05/i; $w[1]=~s/Jun/06/i; $w[1]=~s/Jul/07/i; $w[1]=~s/Aug/08/i;
           $w[1]=~s/Sep/09/i;     $w[1]=~s/O.*/10/i; $w[1]=~s/Nov/11/i; $w[1]=~s/De.*/12/i;

           $r=&date2diff("$w[2].$w[1].$w[4]","$sys{'mday'}.$sys{'mon'}.$sys{'year'}");

           #debug
           # &my_log_id("snapdiff","$s age: $days, hold $t days");

           if ($r < $t) {            # age < hold

               # log
               if ($debug=~/3|h/) { print "    1718: process days: $r\n"; }

               $t=$s;
               $t=~s/_zfs\-..\d*s*_/_zfs_/;  # opt remove old rules
               $t=~s/_zfs_/_zfs-hd$hld\_/;
               if ($s ne $t) {

                     $r=&socket(" rename $s $t",$job{'destination'});
                     if ($r=~/error/) { &my_log_last("rename: $r");  &my_fast_end;  exit; }

                     if ($debug=~/3|h/) { print "      2136 hold days $hld, rename $s\n      to $t\n"; }
               } else {
                     if ($debug=~/3|h/) { print "      2138 hold days &hld $s\n"; }                         # keep the oldest from this year
               }

               return('1');                      # skip delete
           }
       }
  }



  # ok, no extra hold/keep
  if ($debug=~/3|h/) { print "xxxx 1727 destroy snap $s\n"; }
  return('0');
}

 #################  # on initial replication, del dest snap (receiving side) with id
 sub my_del_snap {
 ################# par jobid, nr
 # receiver/my snaps: backup3/nfs.www@1330674968_repli_zfs_backup3_nr_266
 # sender snaps  : www/nfs@1330674968_repli_zfs_backup3_nr_266

   my ($n);
   my $id=$_[0];
   if (!($id=~/^\d+$/))  { print "job-repli delsnap 184: no id $id\n"; return; }

   my $nr=$_[1];     # when nr, delete all but two newest
   if ($nr > 0) {
       $nr=$nr-2;
   }


   my $t=&socket("zfs list -H -r -o name -t snapshot filter:=_repli",$job{'destination_ip'},120);   # cmd, ip, timeout
   if ($r=~/error/) { &my_log_last("list snaps: $t");  &my_fast_end;  exit; }

   my @s=split(/\n/,$t);
   foreach my $s (@s) {

        if (!($s=~/\@$id\_repli/)) { next; }     # not my id
        if (!($s=~/_repli/))       { next; }     # only repli snaps
        print "2199 process $s\n";

        # without nr
        if ($nr eq "") {
              print "2203 zfs destroy $s\n";
              $r=&socket("zfs destroy $s",$job{'destination_ip'},20);
              if ($r=~/error/) {  &my_log_last("del snap: $r");  &my_fast_end;  exit; }
        } else {
               if (!($s=~/_nr_\d+$/))       { next; }     # only nr repli snaps
               $n=$s;
               $n=~s/.*nr\_//;
               if (!($n>0)) { next; }

               if ($n<=$nr) {
                   print "zfs destroy $s\n";                  # keep last two
                   &socket("zfs destroy $s",$job{'destination_ip'},20);
                   if ($r=~/error/) {   &my_log_last("del snap: $t"); &my_fast_end;  exit; }

               }
        }
   }
}


########## end quickly
sub my_fast_end {
##########

      my @p1=caller(0); # this:
      my @p2=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask
      # calling line: $p1[2]

      print "line $p1[2] fast_end\n";

      if (-f "$wpath/_log/jobs/$id.progress")     {  unlink("$wpath/_log/jobs/$id.progress"); }   # progress
      if (-f "$wpath/_log/jobs/$id.running")      {  unlink("$wpath/_log/jobs/$id.running");  }   #
      if (-f "$wpath/_log/jobs/$id.sender")       {  unlink("$wpath/_log/jobs/$id.sender");   }   #
      &my_log_jobs("error in job $id in line $p1[2]");

      if ($id ne "") {
         &my_log_id("error:","fast end in line $p1[2]");
         &my_log_last("error:","fast end in line $p1[2]");
      }
      exit;

}


## cancel job
sub my_cancel  {
###############

   $info="job $id.cancel requested in job-replicate 696";

   # kill all processes with jobid
   # kill receive processes (terminates remote send ?)

    &my_log_last("my cancel, kill all processes");
    &my_fast_end;
}


 ## delete job
 sub my_delete {
 ##############
 my ($r,$t);
      if ($id ne "") {

        &my_cancel;
        if (-f "$wpath/_log/jobs/$id.par")       {  unlink("$wpath/_log/jobs/$id.par");  }   #
        print __LINE__." #### job deleted\n";

      } else {
        print "delete, no id !! ########### \n";
      }

      &my_fast_end;
}

########### log detailed info for last job (klick on Date of last exec)
sub my_log_last {
##########
   &task_gettime;
   my $t=$_[0];
   $t.=" $_[1]";             # like log_id

   my @p0=caller(0);
   my @p1=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask ex $p1[3] l$p1[2]
   my @p2=caller(2);
   my @p3=caller(3);
   my @p4=caller(4);

   ###################################
   # calling line: $p1[2]
   print "$p1[2] $t\n";      # console
   ###################################

                  my $l="$p0[3] line $p0[2]";             $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p1[2] > 0) { $l.="<br> <- $p1[3] line $p1[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p2[2] > 0) { $l.="<br> <- $p2[3] line $p2[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p3[2] > 0) { $l.="<br> <- $p3[3] line $p3[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   if ($p4[2] > 0) { $l.="<br> <- $p4[3] line $p4[2]"; }  $l=~s/\/.*zfsos\// /; $l=~s/\/.*scripts\// /;
   $l=~s/main::/\&/gs;
   $l=~s/br> +/br> /g;
   $l=~s/\s+/ /g;

   $l=~s/\n/<br>/g;
   $t=~s/\n/<br>/gs;


   my $log="$wpath/_log/jobs/$id.last";
   open (PF, ">>$log");
     flock(PF, LOCK_EX);
     print PF "$current{'timestamp'}\t$l\t$t\n";
   close (PF);

   #print __LINE__." $id.last: \n$t\n";   # opt
}


########### write log  to general jobs.log
sub my_log_jobs {
###############
 my (@v,$v,$n,$l,$f1,$f2,@data,$t);
 my $mess=$_[0];

 #$l="ok";
 #if ($mess=~/^error/)  { $l="error"; }

 my @p0=caller(0);
 my @p1=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask ex $p1[3] l$p1[2]
 my @p2=caller(2);
 my @p3=caller(3);
 my @p4=caller(4);

 # calling line: $p1[2]
  $mess=~s/\n/ /gs;

  #stage 1: log general jobs.log
  ###########################

    my $log="$wpath/_log/jobs/jobs.log";
    my $lock="$wpath/_log/jobs/jobs.lock";
    open (X,"<$lock");
          flock(X, LOCK_EX);
          #
          open (F,"<$log");
                flock(F, LOCK_EX);
                @data=<F>;
          close (F);

          #unshift   $l=ok/error
          unshift(@data,"$l; replication $id; time: $current{'timestamp'} line: $p1[2];$mess\n");

          if (@data > 50) {
                     $t=@data-50;
                     splice(@data,50,$t);
          }

         $t=join("",@data);
         open (PF, ">$log");
            flock(PF, LOCK_EX);
            print PF "$t";
         close (PF);
   close (X);
}


###################
sub update_lastrun {
####################
      # update last run in $id.par

      &task_gettime;
      my $log="$wpath/_log/jobs/$id.par";
      open (F,"<$log");
            flock(F, LOCK_EX);
            @data=<F>;
      close (F);
      foreach my $t (@data) {
             if ($t=~/^last=/)   {
                $t="last=$current{'logtime'}\n";
                print "2500 update $t\n";
             }
      }
      $t=join("",@data);

      open (PF, ">$log");
        flock(PF, LOCK_EX);
        print PF "$t";
      close (PF);
      &my_log_last("1522 update id.par: last=$current{'logtime'}");

}


##############  log to id.log
sub my_log_id {
#############

my @p1=caller(0); # this:
my @p2=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask
# calling line: $p1[2]

my $err=$_[0];
my $data="line $p1[2] $_[1]";
$data=~s/ +/ /g;

my @data;
&task_gettime;


  my $log="$wpath/_log/jobs/$id.log";
  open (F,"<$log");
        flock(F, LOCK_EX);
        @data=<F>;
  close (F);

  #unshift   $l=ok/error
  unshift(@data,"$err time: $current{'timestamp'} line $p1[2]; $data\n");

  if (@data > 300) {
             $t=@data-300;
             splice(@data,300,$t);                 # delete from 50+
  }

  my $t=join("",@data);
  open (PF, ">$log");
     flock(PF, LOCK_EX);
     print PF "$t";
  close (PF);

}

 ##################
 sub my_log_timeout {     # set timout as default in id.log, remove when replication ok
 #################

    &my_log_id("check","timeout check - warning is set as default for job not finished correctly, remote timeout or other reason");
    #exit;
 }

########################
sub my_postreplication {
########################
  my ($r,$type,$s,$abe,@data,$t,@t);

  &my_log_last("begin my_postreplication");
  print __LINE__." postreplication\n";

  # rec: set mountpoint to job{'destination'}
  if ($job{'send_rec'}=~/R/)  {
          &my_log_last("rec=$job{'send_rec'},$job{'recursiv'}<br>set mountpoint of $job{'destination_zfs'} to /$job{'destination_zfs'}");
          $r=&socket("zfs set mountpoint=/$job{'destination_zfs'} $job{'destination_zfs'}",$job{'destination_ip'});
          if ($r=~/error/)  {  &my_log_last("zfs set: $r"); &my_fast_end;  exit; }
  }


  # dest fs created?
  $r=&socket("zfs list -r -H -o name -t filesystem $job{'destination_zfs'}",$job{'destination_ip'});     #this one
  if ($r=~/error/)  {   &my_log_last("zfs list: $r"); &my_fast_end;  exit; }
  print __LINE__." zfs list $r\n";

  if ($r ne $job{'destination_zfs'})  {
      print __LINE__." dest fs $job{'destination_zfs'} not created!\n";
      &log_id("error dest fs $job{'destination_zfs'} not created!");
  }

   return;





  #dest writable?       -> not working oon current zfs on win
     #print "\n";
     $r=&socket("zfs list -r -H -o name -t filesystem $job{'destination_zfs'}",$job{'destination_ip'});     #this one and below
     if ($r=~/error/)  {   &my_log_last("zfs list: $r"); &my_fast_end;  exit; }
     my @zfs=split(/\n/,$r);

     foreach my $zfs (@zfs) {
             if ($zfs=~/\/ROOT\//) { next; }                       # BE, do not set readonly
             if ($job{'writable'} =~/enable|yes/) {

               $r=&socket("zfs set readonly=off $zfs",$job{'destination_ip'});
               if ($r=~/error/)  {  &my_log_last("zfs set: $r"); &my_fast_end;  exit; }
               &my_log_last("my postreplication, zfs set readonly=off $zfs\n$r");

             } else {

               $r=&socket("zfs set readonly=on $zfs",$job{'destination_ip'});
               if ($r=~/error/)  { &my_log_last("zfs set: $r"); &my_fast_end;  exit; }
               &my_log_last("my postreplication, zfs set readonly=on $zfs\n$r");


             }
     }



     # opt post job
 #    if (-f "$wpath/_log/jobs/$id.post") {
 #      &my_log_last("execute post shellscript $id.post");
 #      $r=`sudo sh /var/web-gui/_log/jobs/$id.post`;
 #      print "$r\n";
 #      &my_log_last("result of post shellscript\: $r");
 #    }

      # remove ESXi snaps
 #     if ($job{'esxi_ip'} =~/\d/) {
 #       $t=`perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/esxi/job-esxisnap.pl remove_$id`;   # remove ESXi snaps, can happen to fail on some snaps due "too busy"
 #     }
 #
 #    &my_log_last("end my postreplication");



  &my_log_last("begin my_postreplication");

  $r = &grouplib_ask_remote_snaplist("$job{'hostip'}");   #commalist
  @t=split(/,/,$r);

  # debug
  #$r=~s/,/<br>/gs;
  #&my_log_last("snaps on $job{'host'}, next: $job{'next_src_snap'}<br>$r");

  foreach my $t (@t) {
          if ($job{'next_src_snap'} eq "")       { &my_log_last("next_src_snap=n.a.");  last; }
          if (!($t=~/$job{'next_src_snap'}/))    { next; }
          &my_log_last("next_src_snap from host $job{'host'}<br>$t");
  }

  # update last in $id.par
  &task_gettime;
  my $log="/var/web-gui/_log/jobs/$id.par";
  open (F,"<$log");
        flock(F, LOCK_EX);
        @data=<F>;
  close (F);
  foreach my $t (@data) {
         if ($t=~/^last=/)   { $t="last=$current{'logtime'}\n"; }
  }
  $t=join("",@data);
  open (PF, ">$log");
    flock(PF, LOCK_EX);
    print PF "$t";
  close (PF);
  &my_log_last("1522 update id.par: last=$current{'logtime'}");
  ###############################


  # step 1: load destination ZFS and ZFS below
     my $zfs=`$ZFS list -H -o name`;   # all ZFS names
     chomp($zfs);
     my @t=();
     @t=split(/\n/,$zfs);

     #collect ZFS
     my @zfs=();
     foreach my $p (@t) {
            chomp($p);
            if ($p ne $job{'destination'}) { next; }

            $type=`$ZFS list -H -o type`;                     # do not post-process volumes
            chomp($type);
            if ($type eq "volume") { next; }

            push (@zfs,$p);
            #print "found ZFS: $p\n";
     }

  # step 2: set all shares to readonly=off and perm to 777   (if reshared)
     foreach my $zfs (@zfs) {

             $r=`$ZFS set readonly=off \'$zfs\' 2>&1`;
             &my_log_last("zfs set readonly=off $zfs\n$r");

             if ($job{'dest_sharename'} ne "") {         # must be set or may get access denied
                 if ($job{'dest_shareguest'} =~/enable|yes/) {
                      $r= `chmod 777 \'/$zfs\'`;
                      #print "set perm to 777 \'$zfs\'\n$r";
                 } else {
                      $r= `chmod 755 \'/$zfs\'`;
                      #print "set perm to 755 \'$zfs\'\n$r";
                 }
             }
     }

   # step 3 - set all shares to off
     foreach my $zfs (@zfs) {
             $r=  `$ZFS set sharesmb=off \'$zfs\'`;
             $r=  `$ZFS inherit -r sharesmb \'$zfs\'`;      # rec

             $r.= `$ZFS set sharenfs=off \'$zfs\'`;
             $r=  `$ZFS inherit -r sharenfs \'$zfs\'`;      # rec
             &my_log_last("zfs set nfs/smb shares=off $zfs\n$r");
     }
  # step 4 reshare it optionally
     my $cmd;
     if ($job{'dest_sharename'} ne "") {
                  foreach my $zfs (@zfs) {

                         $t="$zfs";
                         $t=~s/\//_/g;  # a/b -> a_b           # sharename=dataset name

                         #overwrite dest_sharename
                                    $s=$job{'destination'};
                                    $s=~s/\//_/g;  # a/b -> a_b

                                    $t=~s/$s/$job{'dest_sharename'}/;

                         #abe?
                         if ($job{'dest_shareabe'} =~/enable|yes/) {
                           $abe=",abe=true";
                         } else {
                           $abe="";
                         }

                         ####
                         if ($sys{'auto_os'} eq "Solaris 11") {

                              if ($job{'dest_shareguest'} =~/enable|yes/) {
                                $r=&exe("$ZFS set share=name=$t,prot=smb$abe,guestok=true $in{'zfs'}");
                              } else {
                                $r=&exe("$ZFS set share=name=$t,prot=smb$abe $in{'zfs'}");

                              }
                              $r=&exe("zfs get -H -o value sharesmb $in{'zfs'}");
                              if ($r eq "off") {           # additional: sharesmb=on
                                  $r=&exe("$sys{'sbinpath'}zfs set sharesmb=on $in{'zfs'}");
                              }

                        } else {

                              ####

                              #$cmd="$ZFS set sharesmb=name=$vars{'dest_sharename'}$t";
                              if ($job{'dest_shareguest'} =~/enable|yes/) {
                                   $r=`$ZFS set sharesmb=name=$t,guestok=true$abe \'$zfs\'`;

                                  # $cmd.=",guestok=true \'$zfs\'; ";
                              } else {
                                   $r=`$ZFS set sharesmb=name=$t$abe \'$zfs\'`;
                                  # $cmd.=" \'$zfs\'; ";
                              }
                         }
                         &my_log_last("zfs set nfs/smb shares=on $zfs\n$r");
                  }
     }

     # step 5 set to read only if writable is not yes
     #print "\n";
     foreach my $zfs (@zfs) {
             if ($zfs=~/\/ROOT\//) { next; }                       # BE, do not set readonly
             if ($job{'writable'} =~/enable|yes/) {

               $r=`$ZFS set readonly=off \'$zfs\' 2>&1`;
               &my_log_last("my postreplication, zfs set readonly=off $zfs\n$r");

             } else {

               $r=`$ZFS set readonly=on \'$zfs\' 2>&1`;
               &my_log_last("my postreplication, zfs set readonly=on $zfs\n$r");
             }
     }

     # opt post job
     if (-f "/var/web-gui/_log/jobs/$id.post") {
       &my_log_last("execute post shellscript $id.post");
       $r=`sudo sh /var/web-gui/_log/jobs/$id.post`;
       print "$r\n";
       &my_log_last("result of post shellscript\: $r");
     }

      # remove ESXi snaps
      if ($job{'esxi_ip'} =~/\d/) {
        $t=`perl /var/web-gui/data/napp-it/zfsos/_lib/scripts/esxi/job-esxisnap.pl remove_$id`;   # remove ESXi snaps, can happen to fail on some snaps due "too busy"
      }

     &my_log_last("end my postreplication");

}

######################
sub my_wait_and_check {  # monitor wait and read cancel/del
#######################
  return;              #   todo

    my $t=$_[0];
    while ($t>0) {
       if (-f "/tmp/nappit/$id.cancel")    { &my_cancel; &my_fast_end;  exit; }
       if (-f "/tmp/nappit/$id.delete")    { &my_delete; exit; }
       sleep (2);
       $t=$t-2;
    }
}


###########
sub my_par {
###########     # read current par_* to %job

    #&task_read_jobpar($id);        # need only current id

    $t=$id."_source_ip";
    $job{'source_ip'}              = $current{$t};                 # senders hostname               ex box3

    $t=$id."_destination_ip";
    $job{'destination_ip'}         = $current{$t};                 # source ZFS                     tank/zfs1

    $t=$id."_source_zfs";
    $job{'source_zfs'}             = $current{$t};                 # dest ZFS                       tank2/zfs1

    $t=$id."_destination_zfs";
    $job{'destination_zfs'}        = $current{$t};                 # dest ZFS                       tank2/zfs1

    $t=$id."_zfs_name";
    $job{'zfs_name'}              = $current{$t};                 # dest ZFS writable             tank2/zfs1

    $t=$id."_writable";
    $job{'writable'}              = $current{$t};                 # dest ZFS writable             tank2/zfs1

    $t=$id."_transport";
    $job{'transport'}              = $current{$t};                 # dest ZFS writable             tank2/zfs1

    $t=$id."_port";
    $job{'port'}              = $current{$t};                 # dest ZFS writable             tank2/zfs1

    #recursiv
    $t=$id."_recursiv";
    if ($current{$t} eq "yes") {
      $t=$id."_snap_rec";
      $job{'snap_rec'}             = "-r";                         # ** snap recursiv               -r or ""

      $t=$id."_send_rec";
      $job{'send_rec'}             = "-R";                         # ** send recursiv               -R or ""
    }

    $t=$id."_send_raw";
    $job{'send_raw'}               = $current{$t};                 # source (sender)                sudo /usr/bin/nc -w 10 172.16.11.28 58309

    $t=$id."_esxi_ip";
    $job{'esxi_ip'}         = $current{$t};                        # destination (receiver)         sudo /usr/bin/nc -d -l -p 58309                 ## no wait !!

    $t=$id."_send_i";
    $job{'send_i'}                 = $current{$t};                  # zfs send -I or -i

    $t=$id."_send_options";
    $job{'send_options'}            = $current{$t};                 # zfs send opt

    $t=$id."_rec_options";
    $job{'rec_options'}              = $current{$t};                # zfs rec opt

    $t=$id."_pv_limit";
    if (!($current{$t}=~/[a-zA-Z0-9]/)) { $current{$t}="";  }
    $job{'pv_limit'}        = $current{$t};                        # throughput limit ex -L 300k

    $t=$id."_keep";
    $job{'keep'}              = $current{$t};                       # keep

    $t=$id."_hold";
    $job{'hold'}              = $current{$t};                       # hold

    ##

}

###### debug
sub print_my_par {
#################
     my ($k,$v,$t);
     #my $id=$_[0];
     $t="job id $id\n";

     #print par
     while (($k,$v) = each %job) {
            $t.= "job $k\=$v\n";
     }
     return ($t);
}

 ###############
 sub my_val2mb  {  #hide:
 ###############

      my $w1=$_[0];
      if ($w1=~/K/) { $w1=~s/K//; $w1=$w1/1000;    }
      if ($w1=~/M/) { $w1=~s/M//; $w1=$w1*1;       }
      if ($w1=~/G/) { $w1=~s/G//; $w1=$w1*1000;    }
      if ($w1=~/T/) { $w1=~s/T//; $w1=$w1*1000000; }
      return ($w1);
 }

 #end
 1;